﻿Imports System.Threading



Public Class frmMain

    ' Sub & Function declarations
    Private attemptsCount As Integer = 0
    Private shortcutOccured As Boolean = False
    Private enableSave As Boolean = False



#Region "Setting Management"
    ''' <summary>
    ''' loadSettings() will load the settings from the memory. The
    ''' settings will be filled in into the form's fields.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub loadSettings()
        With My.Settings
            cmbCOM.SelectedItem = "COM" & .COM
            cmbBAUD.SelectedItem = "" & .BAUD
            inpAddress.Text = .ADDRESS
            chkSoundAlert.Checked = .soundEnabled
        End With
    End Sub




    ''' <summary>
    ''' savesettings() will save the current form fields to the memory
    ''' for later retrieval.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub savesettings()
        If enableSave Then
            With My.Settings
                If cmbCOM.SelectedIndex > -1 Then
                    .COM = cmbCOM.SelectedIndex + 1
                End If

                Dim baudRate As Integer
                If Integer.TryParse(cmbBAUD.Text, baudRate) Then
                    .BAUD = baudRate
                End If

                Dim address As Integer
                If Integer.TryParse(inpAddress.Text, address) Then
                    If address > -1 And address < 256 Then
                        .ADDRESS = address
                    End If
                End If

                .soundEnabled = chkSoundAlert.Checked
            End With
        End If
    End Sub



    ''' <summary>
    ''' frmMain_FormClosing() will save the data settings while quitting the form.
    ''' First the data is validated and the user is given the option to cancel the
    ''' closing in case of incorrect data.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub frmMain_FormClosing(ByVal sender As Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles Me.FormClosing
        If validInputs() = False Then
            Dim ans As Integer
            ans = MsgBox("One or more settings are not valid. You can continue closing this " & _
                         "program but the settings won't be saved. Do you want to cancel the " & _
                         "closing of this program to correct the settings?", _
                         MsgBoxStyle.YesNoCancel + MsgBoxStyle.DefaultButton3 + MsgBoxStyle.Question, _
                         "Close the program?")

            If ans <> MsgBoxResult.No Then
                e.Cancel = True
                Exit Sub
            End If

        Else
            ' Settings are valid. Save them ...
            savesettings()
            My.Settings.Save()
        End If
    End Sub
#End Region




#Region "Communication with the BK"
    ''' <summary>
    ''' buildFrame will construct a data packet with the right size. You need to
    ''' specify the required command (which can be found in the manual of the
    ''' BK1786B), and an optional data byte (byte 4).
    ''' </summary>
    ''' <param name="command">The byte number of the command to execute</param>
    ''' <param name="byte4">Optional data for byte 4. </param>
    ''' <returns>A byte array with 26 positions</returns>
    ''' <remarks></remarks>
    Private Function buildFrame(ByVal command As Byte, Optional ByVal byte4 As Integer = 0) As Byte()
        Dim newFrame(26) As Byte

        ' First we fill the array with zeroes.
        For i As Integer = 0 To 25
            newFrame(i) = 0
        Next

        ' Then we really fill in the actual data
        newFrame(0) = &HAA                  ' A frame always starts with AA
        newFrame(1) = Val(inpAddress.Text)  ' The hardware address
        newFrame(2) = command               ' Setup the command code
        newFrame(3) = byte4

        ' Finally the checksum is calculated and written into the last byte.
        Dim checksum As Integer = calcCheckSum(newFrame)
        newFrame(25) = checksum

        Return newFrame
    End Function




    ''' <summary>
    ''' sendRequest will send a request to the BK1786B. You need to specify
    ''' a data frame with the request, and you will get a data frame in
    ''' return.
    ''' </summary>
    ''' <param name="requestFrame">A byte-array containing the request</param>
    ''' <returns>A byte-array containing the response</returns>
    ''' <remarks></remarks>
    Public Function sendRequest(ByVal requestFrame() As Byte) As Byte()
        Try
            ' Write the frame to the port
            SerialPort1.Write(requestFrame, 0, 26)

            ' Get the number of received bytes and define the buffer
            Dim bufferSize = SerialPort1.BytesToRead
            Dim buffer(bufferSize) As Byte

            ' Loop through all the received data, add it to the buffer
            For i = 0 To bufferSize - 1
                buffer(i) = SerialPort1.ReadByte()
            Next

            ' For debug reasons the respone is added to the Debug window.
            Debug.Print(ByteArrayToString(buffer))

            ' Pass the result to the caller function
            Return buffer
        Catch ex As Exception
            ' Return a single byte with value 0.
            Dim falseResponse(1) As Byte
            falseResponse(0) = 0
            Return falseResponse

        End Try
    End Function




    ''' <summary>
    ''' isDataValid will check if the checksum for a certain data frame
    ''' matches the frame's contents.
    ''' </summary>
    ''' <param name="buffer"></param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function isDataValid(ByVal buffer() As Byte) As Boolean

        Dim valid As Boolean = True

        If buffer.Length = 27 Then
            ' The frame has the expected length of 26 bytes
            Dim checksum As Integer = calcCheckSum(buffer)
            If checksum <> buffer(25) Then
                ' The checksum doesn't match the data
                Debug.Print("invalid data: checksum is wrong")
                valid = False
            End If
        Else
            ' The frame doesn't have the expected length
            valid = False
            Debug.Print("invalid data: frame too short")
        End If

        ' Pass the result
        Return valid
    End Function



    ''' <summary>
    ''' bkCheckConnection() will check if a connection with the BK could be
    ''' established.
    ''' </summary>
    ''' <returns>A true/false indicating if the connection was established</returns>
    ''' <remarks></remarks>
    Public Function bkCheckConnection() As Boolean

        Dim frameToSend() As Byte
        frameToSend = buildFrame(&H31)

        Dim frameReceived() As Byte
        frameReceived = sendRequest(frameToSend)

        If isDataValid(frameReceived) Then
            Return True
        Else
            Return False
        End If
    End Function




    ''' <summary>
    ''' bkCheckOutputs() will check the output states of the BK1786B. It will return a true
    ''' if the device state could be read. Otherwise it will return a false. The data itself
    ''' is passed by the function's parameters.
    ''' </summary>
    ''' <param name="overcurrent">A boolean indicating if the device is in CC (overcurrent) mode.
    ''' You should initially set this boolean to false.</param>
    ''' <param name="state">A boolean indicating the output state of the supply.
    ''' You should initially set this boolean to false.</param>
    ''' <returns>A true/false to indicate if the function could be executed successfully.</returns>
    ''' <remarks></remarks>
    Private Function bkCheckOutputs(ByRef overcurrent As Boolean, ByRef state As Boolean) As Boolean
        Dim frameToSend() As Byte
        frameToSend = buildFrame(&H26)

        Dim frameReceived() As Byte
        frameReceived = sendRequest(frameToSend)

        If isDataValid(frameReceived) Then
            Dim code As Byte = frameReceived(9)

            If (code And 8) = 8 Then
                overcurrent = True
            End If

            If (code And 1) = 1 Then
                state = True
            End If

            Return True
        Else
            ' invalid Frame
            Return False
        End If
    End Function




    ''' <summary>
    ''' calcCheckSum calculates the checksum of a certain data frame. The frame should
    ''' contain 26 bytes, where the 26th byte is the checksum itself. So this last byte 
    ''' is not taken in the calculations.
    ''' </summary>
    ''' <param name="aFrame">The frame you want to calculate the checksum for</param>
    ''' <returns>The calculated checksum. It is determined by the mathematical sum of
    ''' the first 25 bytes, modulus 256.</returns>
    ''' <remarks></remarks>
    Public Function calcCheckSum(ByVal aFrame() As Byte) As Integer
        Dim totSum As Integer = 0

        For i = 0 To 24
            ' TODO: the function should check if the frame has the right length
            ' for now this is not required as the data is checked before it is
            ' passed to the function.
            totSum += aFrame(i)
        Next

        ' Calculate the mathematical checksum and then do a modulus.
        Dim checksum As Integer = totSum Mod 256
        Return checksum

        aFrame(25) = totSum
    End Function

#End Region




#Region "Form Controls"


    ''' <summary>
    ''' frmMain_Load is the sub that is executed when the form is loaded for the first time.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub frmMain_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        lblConnectionStatus.Text = "Not connected"

        ' Fill the COM port field
        For i = 1 To 255
            cmbCOM.Items.Add("COM" & i)
        Next

        ' Fill the baud rate field
        With cmbBAUD.Items
            .Add("4800")
            .Add("9600")
            .Add("19200")
            .Add("38400")
        End With

        loadSettings()
        enableSave = True

    End Sub



    ''' <summary>
    ''' cmbConnect_Click handles the click event for the cmdConnect button.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub cmbConnect_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdConnect.Click

        If SerialPort1.IsOpen Then
            tmrConnectionAttempts.Enabled = False
            tmrReader.Enabled = False
            SerialPort1.Close()
            lblConnectionStatus.Text = "Not connected"
            lblShortCut.Text = "N/A"
            lblOutput.Text = "N/A"
            cmdConnect.Text = "Connect"
        Else
            If validInputs() Then
                ' setup the port
                With SerialPort1
                    .PortName = cmbCOM.Text
                    .Parity = IO.Ports.Parity.None
                    .BaudRate = Val(cmbBAUD.Text)
                    .DataBits = 8
                    .DtrEnable = True
                    .StopBits = IO.Ports.StopBits.One
                End With

                Try
                    SerialPort1.Open()
                    cmdConnect.Text = "Connect"
                    Me.Cursor = Cursors.WaitCursor

                    ' We will now start the timer that attempts to read the BK instrument
                    attemptsCount = 0
                    With tmrConnectionAttempts
                        .Enabled = True
                        .Interval = 1
                    End With

                Catch ex As Exception
                    ' Catch eventual errors
                    MsgBox("An Error occured while connecting: " & ex.Message, _
                            MsgBoxStyle.OkOnly + MsgBoxStyle.Critical, _
                            "Error while connecting")
                End Try


            Else
                MsgBox("Some settings are wrong. Verify them first!", _
                        MsgBoxStyle.OkOnly + MsgBoxStyle.Information, _
                        "Invalid settings")
            End If
        End If

    End Sub




    ''' <summary>
    ''' tmrReader_Tick will poll the state of the BK1786B every 150 milliseconds.
    ''' It will read the output states and let the user hear a sound when the
    ''' supply is in shortcircuit.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tmrReader_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tmrReader.Tick

        Dim overcurrent, status As Boolean
        overcurrent = False
        status = False

        If bkCheckOutputs(overcurrent, status) Then

            If overcurrent Then
                lblShortCut.Text = "YES"
                lblShortCut.ForeColor = Color.Red

                If shortcutOccured = False Then
                    If chkSoundAlert.Checked Then
                        Application.DoEvents()
                        shortcutOccured = True
                        Dim beeper As New Thread(AddressOf beep)
                        beeper.Start()
                    End If
                End If

            Else
                lblShortCut.Text = "NO"
                lblShortCut.ForeColor = Color.Black
                shortcutOccured = False
            End If


            If status Then
                lblOutput.Text = " ON"
            Else
                lblOutput.Text = "OFF"
            End If
        End If
    End Sub




    ''' <summary>
    ''' tmrConnectionAttempts_Tick is launched when the device is connecting for the
    ''' first time. It will check the state of the BK. The tmrReader timer will only
    ''' be started when tmrConnectionAttempts_Tick detects a connection has been made.
    ''' Only 10 connection attempts are made.
    ''' </summary>
    ''' <param name="sender"></param>
    ''' <param name="e"></param>
    ''' <remarks></remarks>
    Private Sub tmrConnectionAttempts_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles tmrConnectionAttempts.Tick
        attemptsCount = attemptsCount + 1
        tmrConnectionAttempts.Interval = 1000

        If bkCheckConnection() = True Then
            ' The timer may start
            tmrConnectionAttempts.Enabled = False
            tmrReader.Enabled = True
            lblConnectionStatus.Text = "Connected!"
            cmdConnect.Text = "Disconnect"
            Me.Cursor = Cursors.Default
        Else
            ' The timer should stop
            If attemptsCount > 5 Then
                tmrConnectionAttempts.Enabled = False
                lblConnectionStatus.Text = "Not connected"
                Me.Cursor = Cursors.Default
                MsgBox("Unable to establish a connection with the BK1786B. Check commununication settings", _
                        MsgBoxStyle.OkOnly + MsgBoxStyle.Information, _
                        "Unable to communicate")
            End If
        End If
    End Sub
#End Region



#Region "Supplementary Code"
    ''' <summary>
    ''' validInputs() will validate the inputs of the form. The function will return true
    ''' if all of the input fields are valid.
    ''' </summary>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Function validInputs() As Boolean
        Dim isValid As Boolean = True
        ErrorProvider1.Clear()

        If cmbCOM.SelectedIndex < 0 Then
            isValid = False
            ErrorProvider1.SetError(cmbCOM, "Select a valid COM port between 1 and 255")

        End If

        If cmbBAUD.SelectedIndex < 0 Then
            isValid = False
            ErrorProvider1.SetError(cmbBAUD, "Select a valid BAUD rate from the list. The default is 4800")
        End If

        Dim address As Integer
        If Integer.TryParse(inpAddress.Text, address) Then
            If address < 0 Or address > 255 Then
                ErrorProvider1.SetError(inpAddress, "Select a valid unique address between 0 and 255")
                isValid = False
            End If
        Else
            ErrorProvider1.SetError(inpAddress, "Select a valid unique address between 0 and 255")
            isValid = False
        End If

        Return isValid
    End Function




    ''' <summary>
    ''' ByteArrayToString() will convert an array with bytes to a hexadecimal String.
    ''' The function will only 
    ''' </summary>
    ''' <param name="byteArray">An array containing the hexadecimal bytes.</param>
    ''' <returns></returns>
    ''' <remarks></remarks>
    Public Function ByteArrayToString(ByVal byteArray() As Byte) As String
        Return BitConverter.ToString(byteArray)
    End Function



    ''' <summary>
    ''' beep() will generate a sound when the BK is in shortcut (CC) mode. The given
    ''' sound lasts for 2 seconds and has a frequency of 4000 Hz.
    ''' </summary>
    ''' <remarks></remarks>
    Private Sub beep()
        System.Console.Beep(4000, 2000)
        shortcutOccured = False
    End Sub
#End Region
End Class
